#include "MyApplication.h"

#include <Wt/Chart/WAbstractChartModel.h>
#include <cfloat>
#include <ranges>
#include <algorithm>

#include <Wt/WBootstrap5Theme.h>

#include <Wt/Chart/WAbstractChartModel.h>
#include <Wt/Chart/WAxisSliderWidget.h>
#include <Wt/Chart/WCartesianChart.h>
#include <Wt/Chart/WDataSeries.h>
#include <Wt/WApplication.h>
#include <Wt/WContainerWidget.h>
#include <Wt/WObject.h>
#include <Wt/WShadow.h>

#include <Wt/WSpinBox.h>
#include <Wt/WDoubleSpinBox.h>
#include <Wt/WLabel.h>
#include <Wt/WPushButton.h>
#include <Wt/WColorPicker.h>
#include <Wt/WCheckBox.h>
#include <Wt/WStandardItemModel.h>

#include <Wt/WPanel.h>


class SinModel : public Chart::WAbstractChartModel
{
public:
    SinModel(vector<Sin>& data, int rows)
        : Chart::WAbstractChartModel{}
        , sinuses{data}
        , rows_{rows}
    { }

    virtual double data(int row, int column) const override {
        double x = -M_PI + row * 2*M_PI / (rowCount() - 1);
        switch ( column ) {
        case 0: return x;
        case 1:
            return ranges::fold_left_first(
                sinuses | views::transform([x](Sin s){return s.get(x);}),
                std::plus<double>()
            ).value_or(0);

        default:
            try { return sinuses.at(column-2).get(x); }
            catch (out_of_range e){ return 0; }
        }
    }
    virtual int columnCount() const override { return 2 + sinuses.size(); };
    virtual int rowCount() const override { return rows_; }

    void setRows(int rows) {
        rows_ = rows;
        changed().emit();
    }

private:
    int rows_;
    vector<Sin>& sinuses;
};


class FourierModel : public Chart::WAbstractChartModel
{
public:
    FourierModel(shared_ptr<SinModel> dm, vector<pair<double, double>>& fd)
        : Chart::WAbstractChartModel{}
        , discreteModel{dm}
        , fourier_data{fd}
    {
        dm->changed().connect([this](){
            fourier_data.clear();
            fourier_data.reserve(rowCount());

            for ( int k = 0; k < rowCount(); k++ ) {
                double x = 0,
                       y = 0;
                for ( int n = 0; n < discreteModel->rowCount(); n++ ) {
                    double e = 2*M_PI*k*n/discreteModel->rowCount();
                    double v = discreteModel->data(n, 1);;

                    x+=v*cos(e);
                    y-=v*sin(e);
                }

                fourier_data.push_back({x, y});
            }

            changed().emit();
        });
    }

    virtual double data(int row, int column) const override {
        switch ( column ) {
        case 0: return row;
        case 1: return sqrt(pow(fourier_data[row].first,2)+pow(fourier_data[row].second,2))/rowCount();
        case 2: return atan2(fourier_data[row].second, fourier_data[row].first);

        default: return 0;
        }
    }
    virtual int columnCount() const override { return 3; };
    virtual int rowCount() const override { return discreteModel->rowCount(); }

private:
    shared_ptr<SinModel> discreteModel;
    vector<pair<double, double>>& fourier_data;
};

class RestoreModel : public Chart::WAbstractChartModel
{
public:
    RestoreModel(vector<pair<double, double>>& fd)
        : fourier_data{fd}
    {}

    virtual double data(int row, int column) const override {
        if ( column == 0 ) return -M_PI + row * 2*M_PI / (rowCount() - 1);;
        if (fourier_data.size()==0) return 0;
        double y = 0;
        for (int k = 0; k < fourier_data.size(); k++) {
            double theta = (2 * M_PI * k * (row+1)) / fourier_data.size();
            y += fourier_data.at(k).first * cos(theta) + fourier_data.at(k).second * sin(theta);
        }
        return y / fourier_data.size();
    }

    virtual int columnCount() const override { return 2; };
    virtual int rowCount() const override { return fourier_data.size(); }

private:
    int rows_ = 0;
    vector<pair<double, double>>& fourier_data;
};


MyApplication::MyApplication(const WEnvironment& env)
    : WApplication{env}
{
    setTheme(make_shared<WBootstrap5Theme>());
    styleSheet().addRule("html, body", "height: 100%");
    root()->addStyleClass("d-flex h-100");

    auto left = root()->addNew<WContainerWidget>();
    left->addStyleClass("d-flex flex-column h-100");

    { // график
        auto chartContainer = left->addNew<WContainerWidget>();
        chart = chartContainer->addNew<Chart::WCartesianChart>();
        //chart->setBackground(WColor(220, 220, 220));
        chart->setType(Chart::ChartType::Scatter);


        roughModel = make_shared<SinModel>(sinuses, 100);
        auto roughSeries = make_unique<Chart::WDataSeries>(1, Chart::SeriesType::Line);
        auto roughSeries_ = roughSeries.get();
        roughSeries_->setModel(roughModel);
        roughSeries_->setXSeriesColumn(0);
        roughSeries->setHidden(true);
        chart->addSeries(std::move(roughSeries));



        detailedModel = make_shared<SinModel>(sinuses, 10000);
        auto seriesPtr = make_unique<Chart::WDataSeries>(1, Chart::SeriesType::Line);
        auto series = seriesPtr.get();
        series->setModel(detailedModel);
        series->setXSeriesColumn(0);
        series->setShadow(WShadow(3, 3, WColor(0, 0, 0, 127), 3));
        chart->addSeries(std::move(seriesPtr));

        chart->setFollowCurve(series);


        discreteModel = make_shared<SinModel>(sinuses, 0);
        auto pointsDSPtr = make_unique<Chart::WDataSeries>(1, Chart::SeriesType::Point);
        pointsDS = pointsDSPtr.get();
        pointsDS->setModel(discreteModel);
        pointsDS->setXSeriesColumn(0);
        pointsDS->setShadow(WShadow(3, 3, WColor(0, 0, 0, 127), 3));
        pointsDS->setHidden(true);
        pointsDS->setLabelsEnabled(Chart::Axis::Y);
        pointsDS->setLabelColor(StandardColor::Black);
        chart->addSeries(std::move(pointsDSPtr));

        restoreModel = make_shared<RestoreModel>(fourier_data);
        auto restoredSignal = make_unique<Chart::WDataSeries>(1, Chart::SeriesType::Curve);
        restoredDS = restoredSignal.get();
        restoredDS->setModel(restoreModel);
        restoredDS->setXSeriesColumn(0);
        restoredDS->setHidden(true);
        restoredDS->setPen({StandardColor::Yellow});
        restoredDS->setShadow(WShadow(3, 3, WColor(0, 0, 0, 127), 3));
        chart->addSeries(std::move(restoredSignal));

        auto restoredSignal_p = make_unique<Chart::WDataSeries>(1, Chart::SeriesType::Point);
        restoredDS_points = restoredSignal_p.get();
        restoredDS_points->setModel(restoreModel);
        restoredDS_points->setXSeriesColumn(0);
        restoredDS_points->setHidden(true);
        restoredDS_points->setPen({StandardColor::Yellow});
        restoredDS_points->setShadow(WShadow(3, 3, WColor(0, 0, 0, 127), 3));
        restoredDS_points->setLabelsEnabled(Chart::Axis::Y);
        restoredDS_points->setLabelColor(StandardColor::Black);
        chart->addSeries(std::move(restoredSignal_p));

        chart->axis(Chart::Axis::X).setMaximumZoomRange( 2*M_PI+1 );
        chart->axis(Chart::Axis::X).setMinimumZoomRange( M_PI / 16.0 );
        chart->axis(Chart::Axis::X).setMinimum( -M_PI );
        chart->axis(Chart::Axis::X).setMaximum(  M_PI );

        chart->axis(Chart::Axis::Y).setMinimumZoomRange(0.1);

        chart->axis(Chart::Axis::X).setLocation(Chart::AxisValue::Zero);
        chart->axis(Chart::Axis::Y).setLocation(Chart::AxisValue::Zero);

        chart->axis(Chart::Axis::X).setGridLinesEnabled(true);
        chart->axis(Chart::Axis::Y).setGridLinesEnabled(true);


        chart->resize(800, 500);

        chart->setPanEnabled(true);
        chart->setZoomEnabled(true);
        chart->setCrosshairEnabled(true);
        chart->setOnDemandLoadingEnabled(true);

        auto sliderWidget = chartContainer->addNew<Chart::WAxisSliderWidget>(roughSeries_);
        sliderWidget->resize(800, 80);
        sliderWidget->setSelectionAreaPadding(40, Side::Left | Side::Right);
    }

    auto controlsContainer = left->addNew<WContainerWidget>();
    controlsContainer->addStyleClass("h-100 d-flex flex-column");
    controlsContainer->setOverflow(Overflow::Hidden);

    { // функции
        auto form = controlsContainer->addNew<WContainerWidget>();
        form->addStyleClass("d-flex flex-row align-items-center gap-1 p-3");

        list = controlsContainer->addNew<WContainerWidget>();
        list->addStyleClass("d-flex flex-column gap-1 p-2");
        list->setOverflow(Overflow::Auto);

        showAll = form->addNew<WCheckBox>();
        showAll->setTristate();
        showAll->clicked().connect([=,this](){
            for ( auto s : sinuses ) {
                ((WCheckBox*)s.el->children().at(0))
                    ->setChecked(showAll->isChecked());
                s.ds->setHidden(!showAll->isChecked());
            }
            checkCount = showAll->isChecked() ? sinuses.size() : 0;
            chart->update();
        });

        form->addNew<WText>("α=");
        auto a = form->addNew<WDoubleSpinBox>();
        a->setValue(1);
        form->addNew<WText>("ω=");
        auto w = form->addNew<WDoubleSpinBox>();
        w->setValue(1);
        form->addNew<WText>("φ=");
        auto f = form->addNew<WDoubleSpinBox>();

        for ( auto i : {a,w,f} ) {
            i->setWidth(100);
            i->setRange(-1000, 1000);
        }

        auto addBtn = form->addNew<WPushButton>("+");

        addBtn->clicked().connect([=,this](){
            addSin({a->value(), w->value(), f->value()});
        });
    }


    auto right = root()->addNew<WContainerWidget>();
    right->addStyleClass("d-flex flex-column h-100 gap-1");
    right->setOverflow(Overflow::Hidden, Orientation::Horizontal);
    { // дискретизация
        auto form = right->addNew<WContainerWidget>();
        form->addStyleClass("d-flex flex-row gap-3 align-items-center my-1");

        form->addNew<WText>("частота дискретизации");
        auto freq = form->addNew<WSpinBox>();
        freq->setWidth(120);
        freq->setRange(0, 1000);
        auto show = form->addNew<WCheckBox>("показать точки");

        show->clicked().connect([=,this](){
            pointsDS->setHidden(!show->isChecked());
            chart->update();
        });


        freq->changed().connect([=,this](){
            discreteModel->setRows(freq->value());
            restoredDS->setHidden( freq->value()==0 || !lineCheck->isChecked() );
            restoredDS_points->setHidden(freq->value()==0 || !pointsCheck->isChecked() );
            chart->update();
        });
    }

    fourierChart = right->addNew<Chart::WCartesianChart>();
    fourierChart->setPlotAreaPadding(100, Side::Right | Side::Left);
    fourierChart->setPlotAreaPadding(50, Side::Top | Side::Bottom);
    fourierChart->setType(Chart::ChartType::Scatter);
    fourierModel = make_shared<FourierModel>(discreteModel, fourier_data);
    fourierModel->changed().connect([this](){restoreModel->changed().emit();});
    fourierChart->setModel(static_pointer_cast<Chart::WAbstractChartModel>(fourierModel));

    auto seriesPtr = make_unique<Chart::WDataSeries>(2, Chart::SeriesType::Point);
    auto phaseSeries = seriesPtr.get();
    phaseSeries->setModel(fourierChart->model());
    phaseSeries->setXSeriesColumn(0);
    phaseSeries->setShadow(WShadow(3, 3, WColor(0, 0, 0, 127), 3));
    fourierChart->addSeries(std::move(seriesPtr));


    seriesPtr = make_unique<Chart::WDataSeries>(1, Chart::SeriesType::Line);
    auto ampSeries = seriesPtr.get();
    ampSeries->setModel(fourierChart->model());
    ampSeries->setXSeriesColumn(0);
    ampSeries->setShadow(WShadow(3, 3, WColor(0, 0, 0, 127), 3));
    fourierChart->addSeries(std::move(seriesPtr));
    auto id = fourierChart->addYAxis(make_unique<Chart::WAxis>());
    ampSeries->bindToYAxis(id);
    fourierChart->setFollowCurve(ampSeries);

    fourierChart->yAxis(0).setTitle("φ");
    fourierChart->yAxis(0).setTextPen(phaseSeries->pen());
    fourierChart->yAxis(id).setTitle("α");
    fourierChart->yAxis(id).setTextPen(ampSeries->pen());

    fourierChart->axis(Chart::Axis::X).setLabelAngle(-90);

    fourierChart->setPanEnabled(true);
    fourierChart->setZoomEnabled(true);
    fourierChart->setCrosshairEnabled(true);

    fourierChart->resize(800, 400);
    fourierChart->axis(Chart::Axis::X).setMinimumZoomRange(0.1);
    auto sliderWidget = right->addNew<Chart::WAxisSliderWidget>(ampSeries);
    sliderWidget->resize(800, 80);
    sliderWidget->setSelectionAreaPadding(40, Side::Left | Side::Right);
    fourierChart->axis(Chart::Axis::X).setZoom(2);
    fourierChart->axis(Chart::Axis::X).setPan(0);


    auto restoreControls = right->addNew<WPanel>();
    restoreControls->setTitle("Восстановленный сигнал");
    restoreControls->setCentralWidget(make_unique<WContainerWidget>());
    auto c = (WContainerWidget*)restoreControls->centralWidget();

    {
        auto line = c->addNew<WContainerWidget>();
        line->addStyleClass("d-flex flex-row align-items-center gap-1 p-3");
        lineCheck = line->addNew<WCheckBox>("показывать график");
        lineCheck->clicked().connect([this](){
            restoredDS->setHidden(!lineCheck->isChecked());
            chart->update();
        });
        auto lineColor = line->addNew<WColorPicker>(WColor{StandardColor::Yellow});
        lineColor->setWidth(50);
        lineColor->changed().connect([lineColor,this](){
            restoredDS->setPen(WPen{lineColor->color()});
            chart->update();
        });
    }
    {
        auto points = c->addNew<WContainerWidget>();
        points->addStyleClass("d-flex flex-row align-items-center gap-1 p-3");
        pointsCheck = points->addNew<WCheckBox>("показывать точки");
        pointsCheck->clicked().connect([this](){
            restoredDS_points->setHidden(!pointsCheck->isChecked());
            chart->update();
        });
        auto pointsColor = points->addNew<WColorPicker>(WColor{StandardColor::Yellow});
        pointsColor->setWidth(50);
        pointsColor->changed().connect([pointsColor,this](){
            WPen p{pointsColor->color()};
            restoredDS_points->setPen(p);
            chart->update();
        });
    }

}

void MyApplication::addSin(Sin s) {
    if ( find(sinuses.cbegin(), sinuses.cend(), s) != sinuses.cend() ) return;

    sinuses.push_back(s);
    int index = sinuses.size()-1;
    roughModel->changed().emit();
    detailedModel->changed().emit();
    discreteModel->changed().emit();

    auto seriesPtr = make_unique<Chart::WDataSeries>(index+2, Chart::SeriesType::Line);
    auto series = seriesPtr.get();
    series->setModel(detailedModel);
    series->setXSeriesColumn(0);
    series->setShadow({3, 3, WColor{0, 0, 0, 127}, 3});
    series->setPen({StandardColor::Green});
    chart->addSeries(std::move(seriesPtr));
    sinuses[index].ds = series;
    series->setHidden(true);

    auto el = list->addNew<WContainerWidget>();
    {
        el->addStyleClass("border rounded-3 p-3 d-flex justify-content-between align-items-center gap-3");

        auto show = el->addNew<WCheckBox>();
        show->clicked().connect([=,this](){
            sinuses.at(series->modelColumn()-2).ds->setHidden(!show->isChecked());
            chart->update();
            checkCount+= show->isChecked() ? 1 : -1;
            showAll->setCheckState(
                checkCount == 0 ?               CheckState::Unchecked
                : checkCount < sinuses.size() ? CheckState::PartiallyChecked
                                              : CheckState::Checked
            );
        });

        auto c = el->addNew<WColorPicker>(WColor{StandardColor::Black});
        c->setColor(StandardColor::Green);
        c->changed().connect([=,this](){
            sinuses.at(series->modelColumn()-2).ds->setPen(WPen{c->color()});
            chart->update();
        });
        c->setWidth(50);

        el->addNew<WText>(WString{"{1} cos( {2}t + {3} )"}.arg(s.a).arg(s.w).arg(s.f));
        auto del = el->addNew<WPushButton>("x");
        del->setStyleClass("btn-danger");

        del->clicked().connect([=,this](){
            removeSin(series->modelColumn()-2);
            el->removeFromParent();
        });
    }

    sinuses[index].el = el;

    int freq = 2*M_PI / ranges::max_element(sinuses, [](Sin a, Sin b){return abs(a.w) > abs(b.w);})->w * 4;
    if ( detailedModel->rowCount() < freq ) detailedModel->setRows(freq);
    else detailedModel->setRows(10000);

    showAll->setCheckState(
        checkCount == 0 ?               CheckState::Unchecked
        : checkCount < sinuses.size() ? CheckState::PartiallyChecked
                                      : CheckState::Checked
    );
}

void MyApplication::removeSin(int index) {
    auto& s = sinuses.at(index);

    chart->removeSeries(s.ds);
    sinuses.erase(next(sinuses.begin(), index));

    for ( ; index < sinuses.size(); index++ )
        sinuses[index].ds->setModelColumn(index+2);

       roughModel->changed().emit();
    detailedModel->changed().emit();
    discreteModel->changed().emit();

    showAll->setCheckState(
        checkCount == 0 ?               CheckState::Unchecked
        : checkCount < sinuses.size() ? CheckState::PartiallyChecked
                                      : CheckState::Checked
    );
}
